#include "../include/dos.h"
void
got_packet(u_char * args, const struct pcap_pkthdr *header,
           const u_char * packet)
{
    eth_hdr *ethernet;          /* The ethernet header */
    ip_hdr *ip;                 /* The IP header */
    u_int id;                   /* Host id */
    u_int i;                    /* Counter */
    struct ifreq ifreq;		/* Interface */
    int tmpfds;			/* Temp file descriptor */
    struct sockaddr_in *sin;	/* n/c */
    char ipaddr[32];		/* Our ip addr */

    static uint32_t packets_read = 0; /* packet counter */
    /* Lock */
    m_packet = 1;

    if (packets_max != 0) {     /* check if -P is specified */
        packets_read++;         /* increase packets_read */
        if (packets_read == packets_max + 1) {  /* check if we hit our counter */
            fprintf(stderr, "Received maximum %u packets\nQuitting...\n",
                    packets_max);
            exit(0);            /* kill ourself politely */
        }
    }

    /* Here we check if packet didn't came from ourselfs */
    tmpfds = socket(AF_INET,SOCK_STREAM,0);
    strncpy(ifreq.ifr_name,opt.pcap_dev,sizeof(ifreq.ifr_name)-1);
    if (ioctl(tmpfds,SIOCGIFADDR,&ifreq)==-1){
      if (errno && errno != ENODEV)
	fprintf(stderr,"ioctl() in got_packet():%s\n",strerror(errno));
      if (errno == ENODEV)
	fprintf(stderr,"Interface %s does not exist!\n",opt.pcap_dev);
    }
    sin = (struct sockaddr_in *)&ifreq.ifr_addr;
    strncpy(ipaddr,inet_ntoa(sin->sin_addr),sizeof(ipaddr));
    close(tmpfds);

    /* Extract ethernet and ip headers from packet */
    ethernet = (eth_hdr *) (packet);  /* Pointer to ethernet header */
    if (strncmp(opt.pcap_dev,"ppp",3) == 0)	/* problem with ppp header */
        ip = (ip_hdr *) (packet + sizeof(eth_hdr)-2+PPP_HDRLEN); /* if ppp header then set offset */
    else
        ip = (ip_hdr *) (packet + sizeof(eth_hdr)); /* Pointer to IP header */

    if (strncmp(ipaddr,inet_ntoa(ip->ip_src),strlen(ipaddr))==0)
      if (strlen(ipaddr)==strlen(inet_ntoa(ip->ip_src))) /* if packet from */
	return;			/* our IP then stop */

    if (verbose > 1){
	printf("Got packet: %s->",inet_ntoa(ip->ip_src));
	printf("%s\n",inet_ntoa(ip->ip_dst));
    }

	/* Check netfilters */
	if (netfilter_num > 0) {
		uint8_t	hit = 0;
		i = 0;

		while (i < netfilter_num) {
			if ((ntohl(ip->ip_src.s_addr) & netfilter[i].mask) == ntohl(netfilter[i].addr.s_addr))
				hit = 1;
			i++;
		}

		/* IP dont match filtered net */
		if (!hit)
			return;
	}

    /* Add host to tree */
    if ((id = find_host(1, ip->ip_src)) == 0)
        id = add_host(1, ip->ip_src, 1);

    /* Match rules */
    for (i = 0; i < rules_num; i++) {
        if (match_packet(packet, &rules[i]) == 0) {
            /* Score host */
	    if (verbose > 0)
		printf("Scoring host: %s\n\n",inet_ntoa(ip->ip_src));
            score_host(id, rules[i].score);
            /* Stop on quick rule */
            if (rules[i].type == TYPE_QUICK)
                break;
        }
    }

    /* Unlock */
    m_packet = 0;

    /* Run display() if needed */
    if (m_display) {
#ifdef DEBUG
        printf("Coming back to display()\n");
#endif
        display(0);
    }
}
