#include "../include/dos.h"

void display(int sig)
{

    uint32_t i;						/* Counter */
    time_t result;					/* Time structure */
    char *t = "";					/* Time in string format */
    static uint32_t run_time = 0 ;	/* Runtime counter */

	/* Increase runtime */
    run_time++;

    /* Get timestamp if needed */
    if (run_time % opt.refresh == 0) {  /* if we hit our refresh time */
        if (opt.timestamp) {
            result = time(NULL);
            t = asctime(localtime(&result));
            t[strlen(t) - 1] = ' ';
            t[strlen(t)] = 0;
        }

        /* Check mutex */
        if (m_packet) {
#ifdef DEBUG
            printf("Interrupt got_packet()\n");
#endif
            m_display = 1;
            return;
        }

        /* Lock mutex for display() */
        m_display = 1;

#ifdef DEBUG_DISPLAY
        printf("Processing display() routine\n");
#endif

        /* Make output to logfile */
        if (opt.logfile)
            log_file(t);

        /* Make output to screen */
        if (!opt.daemon)
            for (i = 0; i <= max_id; i++) {
				/* Quiet mode */
				if (opt.quiet) {
                	if (hosts[i].pts >= opt.critical)
						/* Redirector output */
						if (opt.redir) {
							switch(opt.rtype) {
								case REDIR_IPNAT:
										printf("rdr %s from %s/32 to 0.0.0.0/0 port = 80 -> %s port %d\n", opt.pcap_dev, hosts[i].ipstr, opt.rhost, opt.rport);
										break;
							}
						} else printf("%s\n", hosts[i].ipstr);
				} else  {
					/* Normal output */
                	if (hosts[i].pts >= opt.critical)
                   	 printf("%sCRITICAL : %s got %d points.\n", t,
                   	        hosts[i].ipstr, hosts[i].pts);
                	else if (hosts[i].pts >= opt.warning)
                   	 printf("%sWarning : %s got %d points.\n", t,
                   	        hosts[i].ipstr, hosts[i].pts);
				}
            }

        /* Reset points counter */
        for (i = 0; i <= max_id; i++)
            hosts[i].pts = 0;

        /* Unlock mutex */
        m_display = 0;
    }
    if (opt.timeout <= run_time) {  /* it's time to die */
        fprintf(stderr,
                "Time to run (%u seconds) exceeded - Quitting...\n",
                run_time);
        exit(0);
    }
}
